"""
Contents: Routines to get data from the spotter api

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines to get data from the spotter search api

Functions:

- `search_circle`, get all available data within a given circle.
- `search_rectangle`, get all available data within a given rectangle.

"""
# 1) Imports
# =============================================================================
from datetime import datetime
from pysofar.spotter import SofarApi
from roguewave.tools.time import datetime_to_iso_time_string
from roguewave.wavespectra import concatenate_spectra

from typing import (
    Union,
    Tuple,
    Literal,
    Sequence,
)
from .helper_functions import _get_sofar_api, _get_class, _unique_filter, as_dataframe
import roguewave.spotterapi.spotter_cache as spotter_cache

# 2) Constants & Private Variables
# =============================================================================


DATA_TYPES = Literal[
    "waves", "wind", "surfaceTemp", "barometerData", "frequencyData", "microphoneData"
]


# 3) Main Functions
# =============================================================================
def search_circle(
    start_date: Union[datetime, str],
    end_date: Union[datetime, str],
    center_lat_lon: Tuple,
    radius: float,
    session: SofarApi = None,
    cache=True,
):
    """
    Search for all Spotters that have data available within the give spatio-
    temporal region defined by a circle with given center and radius and start-
    and end- dates. This calls the "search" endpoint of wavefleet.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
    :param center_lat_lon: (Latitude, Longitude) of the center of the circle.
    :param radius: Radius in meter of the circle.
    :param session: Active SofarApi session. If none is provided one will be
                    created automatically. This requires that an API key is
                    set in the environment.
    :return:
    """

    geometry = {"type": "circle", "points": center_lat_lon, "radius": radius}

    if session is None:
        session = _get_sofar_api()

    if cache:
        return spotter_cache.get_data_search(
            handler=_search,
            session=session,
            start_date=start_date,
            end_date=end_date,
            geometry=geometry,
        )
    else:
        #
        return _search(start_date, end_date, geometry, session)


def search_rectangle(
    start_date: Union[datetime, str],
    end_date: Union[datetime, str],
    bounding_box,
    session: SofarApi = None,
    cache=True,
):
    """
    Search for all Spotters that have data available within the give spatio-
    temporal region defined by a circle with given center and radius and start-
    and end-dates. This calls the "search" endpoint of wavefleet.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
    :param bounding_box: coordinates of two points that define a rectangular
    bounding box. Coordinates per point are given as (lat, lon) pairs, and the
    input takes the form of a list/tuple of points: ( (p1_lat, p1_lon),(p2_lat,
    p2_lon) )
    :param session: Active SofarApi session. If none is provided one will be
                    created automatically. This requires that an API key is
                    set in the environment.
    :return:
    """
    geometry = {"type": "envelope", "points": bounding_box, "radius": None}

    if session is None:
        session = _get_sofar_api()

    print("Get Spotter data: retrieving all data from spatio-temporal region")
    if cache:
        return spotter_cache.get_data_search(
            _search,
            session,
            start_date=start_date,
            end_date=end_date,
            geometry=geometry,
        )
    else:
        #
        return _search(start_date, end_date, geometry, session)


# 5) Private Functions
# =============================================================================
def _search(
    start_date: Union[datetime, str],
    end_date: Union[datetime, str],
    geometry: dict,
    session: SofarApi = None,
    variables: Sequence[DATA_TYPES] = None,
    page_size=500,
):
    if session is None:
        session = _get_sofar_api()

    if variables is None:
        variables = ["waves", "wind", "surfaceTemp", "barometerData", "frequencyData"]

    start_date_str = datetime_to_iso_time_string(start_date)
    end_date_str = datetime_to_iso_time_string(end_date)

    shape = geometry["type"]
    shape_params = geometry["points"]
    radius = geometry["radius"]

    generator = session.search(
        shape=shape,
        shape_params=shape_params,
        start_date=start_date_str,
        end_date=end_date_str,
        radius=radius,
        page_size=page_size,
        return_generator=True,
    )

    spotters = {}
    # loop over all spotters returned
    for spotter in generator:
        spotter_id = spotter["spotterId"]
        # loop over keys we can parse
        for key in variables:
            #

            if key in spotter:
                item = spotter[key]
                if not item:
                    # no data
                    continue

                item["latitude"] = spotter["latitude"]
                item["longitude"] = spotter["longitude"]
                item["timestamp"] = spotter["timestamp"]
                data = _get_class(key, item)
                if spotter_id not in spotters:
                    spotters[spotter_id] = {}

                if key not in spotters[spotter_id]:
                    spotters[spotter_id][key] = []

                spotters[spotter_id][key].append(data)

    for spotter_id in spotters:
        for key in spotters[spotter_id]:
            # ensure results are unique
            unique = _unique_filter(spotters[spotter_id][key])

            if key == "frequencyData":
                spotters[spotter_id][key] = concatenate_spectra(unique, dim="time")
            else:
                spotters[spotter_id][key] = as_dataframe(unique)

    return spotters



from roguewave import FrequencySpectrum, FrequencyDirectionSpectrum, WaveSpectrum
from numpy.random import default_rng
from numpy.fft import irfft
from numpy import sqrt, cos, sin, pi, exp, sum, linspace
from numpy.typing import NDArray
from typing import Literal, Tuple


def surface_timeseries(
    component: Literal["u", "v", "w", "x", "y", "z"],
    sampling_frequency: float,
    signal_length: int,
    spectrum: WaveSpectrum,
    seed: int = None,
) -> Tuple[NDArray, NDArray]:
    """
    Create a timeseries for from a given power spectral density.

    :param component: Wave component to create a timeseries for: u,v,w,x,y,z.
    :param sampling_frequency: Sampling frequency of output signal in Hertz
    :param signal_length: Length of output signal
    :param spectrum: Input power spectrum
    :param seed: Input seed for the random number generator.
    :return:
    """

    nfft = (int(signal_length) // 2) * 2

    frequencies = linspace(0, 0.5 * sampling_frequency, nfft // 2, endpoint=False)

    time = linspace(0, nfft / sampling_frequency, nfft, endpoint=False)

    timeseries = nfft * irfft(
        create_fourier_amplitudes(component, spectrum, frequencies, seed)
    )

    return time, timeseries


def create_fourier_amplitudes(
    component, spectrum: WaveSpectrum, frequencies, seed=None
):
    spectrum = spectrum.interpolate_frequency(frequencies, extrapolate_tail=True)

    if isinstance(spectrum, FrequencySpectrum):
        radian_directions = 0.0
        radian_frequency = spectrum.radian_frequency.values
        area = spectrum.frequency_step.values

    elif isinstance(spectrum, FrequencyDirectionSpectrum):
        radian_directions = spectrum.radian_direction.values[None, :]
        radian_frequency = spectrum.radian_frequency.values[:, None]
        area = (
            spectrum.frequency_step.values[:, None]
            * spectrum.direction_step.values[None, :]
        )
    else:
        raise ValueError("Not a spectrum")

    if component == "w":
        factor = 1j * radian_frequency

    elif component == "u":
        factor = radian_frequency * cos(radian_directions)

    elif component == "v":
        factor = radian_frequency * sin(radian_directions)

    elif component == "x":
        factor = -1j * cos(radian_directions)

    elif component == "y":
        factor = -1j * sin(radian_directions)

    else:
        factor = 1.0

    phases = default_rng(seed=seed).uniform(0, 2 * pi, spectrum.spectral_shape())
    amplitudes = sqrt(area * spectrum.variance_density / 2) * exp(1j * phases) * factor

    if isinstance(spectrum, FrequencyDirectionSpectrum):
        amplitudes = sum(amplitudes, axis=-1)

    return amplitudes



"""

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Smit
"""
import os
import dotenv
import requests
import json
from pysofar import SofarConnection
from pysofar.wavefleet_exceptions import QueryError
from functools import cached_property
from roguewave.wavespectra import (
    FrequencyDirectionSpectrum,
    create_2d_spectrum,
    concatenate_spectra,
)
import typing
import netCDF4
import numpy
from typing import List, overload, Union, Dict
from roguewave.tools.time import to_datetime64


def get_token():
    # config values
    userpath = os.path.expanduser("~")
    enviromentFile = os.path.join(userpath, "sofar_api.env")
    dotenv.load_dotenv(enviromentFile)
    token = os.getenv("SPECTRAL_API_TOKEN")
    _wavefleet_token = token

    return _wavefleet_token


def get_endpoint():
    _endpoint = os.getenv("SPECTRAL_URL")
    if _endpoint is None:
        _endpoint = "https://api.sofarocean.com/api"
    return _endpoint


class SofarSpectralAPI(SofarConnection):
    def __init__(self, costum_token=None):
        #
        token = costum_token if costum_token else get_token()
        super().__init__(token)
        self.endpoint = get_endpoint()

    @cached_property
    def _query(self):
        scode, data = self._get("op-wave-spectra")

        if scode != 200:
            raise QueryError(data["message"])

        return data["data"]

    def _download(self, latitude, longitude):
        # Helper methods to download files

        url = f"{self.endpoint}/op-wave-spectra/{latitude}/{longitude}"
        response = requests.get(url, headers=self.header)

        status = response.status_code
        if status != 200:
            data = json.loads(response.text)
            raise QueryError(data["message"])

        data = response.content
        return data

    def points(self):
        data = self._query
        return [{"latitude": x["latitude"], "longitude": x["longitude"]} for x in data]

    def urls(self):
        data = self._query
        return [x["url"] for x in data]

    def download_spectral_file(self, latitude, longitude, directory) -> str:
        os.makedirs(directory, exist_ok=True)
        filename = f"spectrum_{longitude}E_{latitude}N.nc"
        filepath = os.path.join(directory, filename)

        with open(filepath, "wb") as file:
            file.write(self._download(latitude, longitude))

        return filepath

    def download_all_spectral_files(self, directory) -> typing.List[str]:
        points = self.points()
        return [
            self.download_spectral_file(
                point["latitude"], point["longitude"], directory
            )
            for point in points
        ]


@overload
def load_sofar_spectral_file(filename: str) -> FrequencyDirectionSpectrum:
    ...


@overload
def load_sofar_spectral_file(filename: List[str]) -> List[FrequencyDirectionSpectrum]:
    ...


@overload
def load_sofar_spectral_file(
    filename: Dict[str, str]
) -> Dict[str, FrequencyDirectionSpectrum]:
    ...


def load_sofar_spectral_file(filename: Union[str, List[str], Dict[str, str]]):

    if isinstance(filename, str):
        if not os.path.exists(filename):
            raise FileNotFoundError()

        dataset = netCDF4.Dataset(filename)

        nt = dataset.dimensions["time"].size
        frequencies = dataset.variables["frequencies"][:]
        directions = dataset.variables["directions"][:]

        spectra = []
        for ii in range(0, nt):
            unixepochtime = dataset.variables["time"][ii]
            latitude = dataset.variables["latitude"][0]
            longitude = dataset.variables["longitude"][0]
            variance_density = numpy.squeeze(
                dataset.variables["frequency_direction_spectrum"][ii, :, :]
            )

            if (
                dataset.variables["frequency_direction_spectrum"].dimensions[1]
                == "directions"
            ):
                variance_density = variance_density.transpose()

            spectra.append(
                create_2d_spectrum(
                    frequency=frequencies,
                    variance_density=variance_density,
                    time=to_datetime64(unixepochtime),
                    latitude=latitude,
                    longitude=longitude,
                    direction=directions,
                )
            )
        return concatenate_spectra(spectra, dim="time")
    elif isinstance(filename, dict):
        out = {}
        for key in filename:
            out[key] = load_sofar_spectral_file(filename[key])
        return out
    elif isinstance(filename, List):
        return [load_sofar_spectral_file(name) for name in filename]
    else:
        raise Exception("Unsupported input type for filename")



import s3fs
import numpy
from multiprocessing.pool import Pool
from itertools import repeat
import tqdm
from pandas import DataFrame
from roguewave.tools.time import to_datetime64, to_datetime_utc
from roguewave.interpolate.cluster import interpolate_cluster
from xarray import DataArray, open_dataset
from roguewave.interpolate.geometry import ClusterStack, Cluster, Geometry, \
    convert_to_track_set, convert_to_cluster_stack
from typing import Dict,List, Union
from roguewave.modeldata.modelinformation import _get_resource_specification
from roguewave.modeldata.keygeneration import generate_uris
BLOCKSIZE = 1024

def extract_clusters(geometry: Geometry,
                     variable,
                     time_slice,
                     model_name,
                     parallell=False,
                     number_of_workers=10) -> Dict[str, DataFrame]:


    resource = _get_resource_specification(model_name)
    time_base = time_slice.time_base(resource.model_time_configuration)
    valid_time = to_datetime64([ x[0]+x[1] for x in time_base ])

    clusters = convert_to_cluster_stack(geometry, valid_time)

    if isinstance(variable,str):
        variable = [variable]

    out = {  }
    for var in variable:
        uris = generate_uris(var, time_slice, model_name)
        data = _extract_clusters(clusters,uris,var,parallell,
                                 number_of_workers)
        for name, dataframe in data.items():
            if name not in out:
                out[name] = dataframe
            else:
                out[name][var] = dataframe[var]
    return out


def _extract_clusters(clusters: ClusterStack,
                     uris, variable,
                     parallell=False,
                     number_of_workers=10) -> Dict[str, DataFrame]:
    filesystem = s3fs.S3FileSystem()

    # Some properties are assumed homogeneous acrros the files. We can avoid
    # rereading them by caching from the first entry
    with filesystem.open(uris[0], 'rb', cache_type='bytes',
                         block_size=BLOCKSIZE) as file:

        with open_dataset(file, engine='h5netcdf') as dataset:
            latitude = dataset['latitude'].values
            longitude = dataset['longitude'].values

    # Create iterable that serves as input to the worker
    work = zip(clusters.clusters,
               repeat(latitude),
               repeat(longitude),
               repeat(variable),
               uris)

    if parallell:
        with Pool(processes=number_of_workers) as pool:
            data = list(
                tqdm.tqdm(pool.imap(worker, work),
                total=len(clusters))
            )
    else:
        data = list(tqdm.tqdm(map(worker, work)))

    # Transform to a dictionary, with point names as keys and all the relevant
    # values as items
    out = {}
    for index, cluster_data in enumerate(data):
        for point_name, point_value in cluster_data.items():
            if point_name not in out:
                out[point_name] = {'time': [], variable: []}

            out[point_name]['time'].append(
                to_datetime_utc(clusters.time[index]))
            out[point_name][variable].append(point_value)

    _out = {}
    for point_name, point_data in out.items():
        _out[point_name] = DataFrame(index=point_data['time'])
        _out[point_name][variable] = numpy.array(point_data[variable])
    return _out


def worker(arg) -> Dict[str, float]:
    cluster = arg[0]  # type:Cluster
    latitudes = arg[1]
    longitudes = arg[2]
    variable = arg[3]
    uri = arg[4]
    filesystem = s3fs.S3FileSystem()
    out = {}

    for name, point in cluster.points.items():
        if (point.is_valid):
            break
    else:
        # No valid points
        out = {}
        for key in cluster.points:
            out[key] = numpy.nan
        return out

    with filesystem.open(uri, 'rb', cache_type='bytes',
                         blocksize=BLOCKSIZE) as file:

        with open_dataset(file, engine='h5netcdf') as ds:
            scale_factor = get_attr_netcdf(ds.variables[variable].attrs,
                                           'scale_factor', 1)
            FillValue = get_attr_netcdf(ds.variables[variable].attrs,
                                        '_FillValue', numpy.nan)
            missing_value = get_attr_netcdf(ds.variables[variable].attrs,
                                            '_missing_value', numpy.nan)
            add_offset = get_attr_netcdf(ds.variables[variable].attrs,
                                         'add_offset', 0)

            def get_data(indices, _dummy):
                value = ds[variable][
                    ..., DataArray(indices[0]), DataArray(indices[1])].values
                value = numpy.squeeze(value)
                mask = (value == FillValue) | (value == missing_value)
                value[mask] = numpy.nan
                return value * scale_factor + add_offset

            out = interpolate_cluster(
                latitudes,
                longitudes,
                cluster,
                get_data
            )
    return out


def get_attr_netcdf(attrs, name, default):
    # Not all netcdf variable encode all the attributes we are looking for.
    # Here we set default values in case they are missing
    try:
        return attrs[name]
    except:
        return default


import numpy
from roguewave.wavespectra.estimators.mem2 import mem2
from roguewave.wavespectra.estimators.mem import mem
from numba_progress import ProgressBar
from typing import Literal

Estimators = Literal["mem", "mem2"]


# -----------------------------------------------------------------------------
#                       Boilerplate Interfaces
# -----------------------------------------------------------------------------
def estimate_directional_spectrum_from_moments(
    e: numpy.ndarray,
    a1: numpy.ndarray,
    b1: numpy.ndarray,
    a2: numpy.ndarray,
    b2: numpy.ndarray,
    direction: numpy.ndarray,
    method: Estimators = "mem2",
    **kwargs,
) -> numpy.ndarray:
    """
    Construct a 2D directional distribution based on the directional moments and a spectral
    reconstruction method.

    :param number_of_directions: length of the directional vector for the
    2D spectrum. Directions returned are in degrees

    :param method: Choose a method in ['mem','mem2']
        mem: maximum entrophy (in the Boltzmann sense) method
        Lygre, A., & Krogstad, H. E. (1986). Explicit expression and
        fast but tends to create narrow spectra anderroneous secondary peaks.

        mem2: use entrophy (in the Shannon sense) to maximize. Likely
        best method see- Benoit, M. (1993).

    REFERENCES:
    Benoit, M. (1993). Practical comparative performance survey of methods
        used for estimating directional wave spectra from heave-pitch-roll data.
        In Coastal Engineering 1992 (pp. 62-75).

    Lygre, A., & Krogstad, H. E. (1986). Maximum entropy estimation of the
        directional distribution in ocean wave spectra.
        Journal of Physical Oceanography, 16(12), 2052-2060.

    """
    return estimate_directional_distribution(a1,b1,a2,b2,direction, method,**kwargs) * e[...,None]


def estimate_directional_distribution(
    a1: numpy.ndarray,
    b1: numpy.ndarray,
    a2: numpy.ndarray,
    b2: numpy.ndarray,
    direction: numpy.ndarray,
    method: Estimators = "mem2",
    **kwargs,
) -> numpy.ndarray:
    """
    Construct a 2D directional distribution based on the directional moments and a spectral
    reconstruction method.

    :param number_of_directions: length of the directional vector for the
    2D spectrum. Directions returned are in degrees

    :param method: Choose a method in ['mem','mem2']
        mem: maximum entrophy (in the Boltzmann sense) method
        Lygre, A., & Krogstad, H. E. (1986). Explicit expression and
        fast but tends to create narrow spectra anderroneous secondary peaks.

        mem2: use entrophy (in the Shannon sense) to maximize. Likely
        best method see- Benoit, M. (1993).

    REFERENCES:
    Benoit, M. (1993). Practical comparative performance survey of methods
        used for estimating directional wave spectra from heave-pitch-roll data.
        In Coastal Engineering 1992 (pp. 62-75).

    Lygre, A., & Krogstad, H. E. (1986). Maximum entropy estimation of the
        directional distribution in ocean wave spectra.
        Journal of Physical Oceanography, 16(12), 2052-2060.

    """

    # Jacobian to transform distribution as function of radian angles into
    # degrees.
    Jacobian = numpy.pi / 180
    direction_radians = direction * Jacobian

    if method.lower() in ["maximum_entropy_method", "mem"]:
        # reconstruct the directional distribution using the maximum entropy
        # method.
        function = mem
    elif method.lower() in ["maximum_entrophy_method2", "mem2"]:
        function = mem2
    else:
        raise Exception(f"unsupported spectral estimator method: {method}")

    output_shape = list(a1.shape) + [len(direction)]
    if a1.ndim == 1:
        input_shape = [1, a1.shape[-1]]
    else:
        input_shape = [numpy.prod(a1.shape[0:-1]), a1.shape[-1]]

    a1 = a1.reshape(input_shape)
    b1 = b1.reshape(input_shape)
    a2 = a2.reshape(input_shape)
    b2 = b2.reshape(input_shape)

    number_of_iterations = a1.shape[0]
    if number_of_iterations < 100:
        disable = True
    else:
        disable = False

    if method != "mem2":
        msg = f"Reconstructing 2d spectrum with {method} using implementation: "
    else:
        solution_method = kwargs.get("solution_method", "scipy")
        msg = f"Reconstructing 2d spectrum with {method} using solution_method {solution_method}"

    with ProgressBar(total=number_of_iterations, disable=disable, desc=msg) as progress:
        res = function(direction_radians, a1, b1, a2, b2, progress, **kwargs)

    return res.reshape(output_shape) * Jacobian



"""
Implementation of the "MEM2" method:

see Kim1995:

    Kim, T., Lin, L. H., & Wang, H. (1995). Application of maximum entropy method
    to the real sea data. In Coastal Engineering 1994 (pp. 340-355).

    link: https://icce-ojs-tamu.tdl.org/icce/index.php/icce/article/download/4967/4647
    (working as of May 29, 2022)

and references therein.

"""
import numpy
from scipy.optimize import root
import typing
from roguewave.wavespectra.estimators.utils import get_direction_increment
from roguewave.wavespectra.estimators.mem import numba_mem
from numba import njit, prange
from numba.typed import Dict as NumbaDict
from numba.core import types
from numpy.linalg import norm
from numba_progress import ProgressBar

# Settings for numba JIT compilation- whether to use fast math and parallel optimizations when possible.
_FASTMATH = True
_PARALLEL = False

# Numerical settings used in solving for the mem2 distribution
NUMERICS = {
    # absolute tolerence stopping criterium. let moment = [ a1,b1,a2,b2] and let iterate_moment contain the moments
    # calculated from the current estmitaed distribution. The stopping criterium is:
    #     norm( moment-iterate_moment ) < atol
    "atol": 0.01,
    # Maximum number of iterations
    "max_iter": 100,
    # Maximum number of subiterations in the line search algorithm. Typically deep line search activates only when
    # the convergence is poor anyway.
    "max_line_search_depth": 8,
    # If we fall back to least squares estimate of the newton update we have an ill-conditioned system, and solve
    # the system approximately removing the smallest singular values. rcond it the ration of smallest divided by largest
    # singular value.
    "rcond": 1e-6,
    # Convergence is mostly (based on limited testing) poor for narrow distributions (large lagrange multipliers). If
    # we fail to converge we fall back to the mem estimate which has no such issues. For narrow distributions this is
    # hopefully fine.
    "use_mem_when_failing_to_converge": True,
}

# Entry Function
# =============================================================================


def mem2(
    directions_radians: numpy.ndarray,
    a1: numpy.ndarray,
    b1: numpy.ndarray,
    a2: numpy.ndarray,
    b2: numpy.ndarray,
    progress_bar: ProgressBar = None,
    solution_method="newton",
    solver_config=None,
) -> numpy.ndarray:
    """

    :param directions_radians:
    :param a1:
    :param b1:
    :param a2:
    :param b2:
    :param solution_method:
    :return:
    """

    if solver_config is None:
        solver_config = NUMERICS

    else:
        solver_config = NUMERICS | solver_config

    if solution_method == "scipy":
        func = mem2_scipy_root_finder
        kwargs = {}

    elif solution_method == "newton":
        func = mem2_newton
        numba_solver_config = NumbaDict.empty(
            key_type=types.unicode_type, value_type=types.float64
        )
        for key in solver_config:
            numba_solver_config[key] = solver_config[key]

        kwargs = {"config": numba_solver_config}

    elif solution_method == "approximate":
        func = mem2_newton
        kwargs = {"approximate": True}

    else:
        raise ValueError("Unknown method")

    return func(directions_radians, a1, b1, a2, b2, progress_bar, **kwargs)


# Scipy Implementation
# =============================================================================
def mem2_scipy_root_finder(
    directions_radians: numpy.ndarray,
    a1: typing.Union[numpy.ndarray, float],
    b1: typing.Union[numpy.ndarray, float],
    a2: typing.Union[numpy.ndarray, float],
    b2: typing.Union[numpy.ndarray, float],
    progress,
    **kwargs
) -> numpy.ndarray:
    """
    Return the directional distribution that maximizes Shannon [ - D log(D) ]
    enthrophy constrained by given observed directional moments,

    :param directions_radians: 1d array of wave directions in radians,
    length[number_of_directions]

    :param a1: 1d array of cosine directional moment as function of frequency,
    length [number_of_frequencies]

    :param b1: 1d array of sine directional moment as function of frequency,
    length [number_of_frequencies]

    :param a2: 1d array of double angle cosine directional moment as function
    of frequency, length [number_of_frequencies]

    :param b2: 1d array of double angle sine directional moment as function of
    frequency, length [number_of_frequencies]

    :return: array with shape [number_of_frequencies,number_of_direction]
    representing the directional distribution of the waves at each frequency.

    Maximize the enthrophy of the solution with entrophy defined as:

           integrate - D * log(D) over directions

    such that the resulting distribution D reproduces the observed moments.

    """

    number_of_frequencies = a1.shape[-1]
    number_of_points = a1.shape[0]

    directional_distribution = numpy.zeros(
        (number_of_points, number_of_frequencies, len(directions_radians))
    )

    direction_increment = get_direction_increment(directions_radians)

    twiddle_factors = numpy.empty((4, len(directions_radians)))
    twiddle_factors[0, :] = numpy.cos(directions_radians)
    twiddle_factors[1, :] = numpy.sin(directions_radians)
    twiddle_factors[2, :] = numpy.cos(2 * directions_radians)
    twiddle_factors[3, :] = numpy.sin(2 * directions_radians)

    guess = initial_value(a1, b1, a2, b2)
    for ipoint in range(0, number_of_points):
        progress.update(1)
        for ifreq in range(0, number_of_frequencies):
            #
            moments = numpy.array(
                [
                    a1[ipoint, ifreq],
                    b1[ipoint, ifreq],
                    a2[ipoint, ifreq],
                    b2[ipoint, ifreq],
                ]
            )

            if numpy.any(numpy.isnan(guess[ipoint, ifreq, :])):
                continue

            res = root(
                moment_constraints,
                guess[ipoint, ifreq, :],
                args=(twiddle_factors, moments, direction_increment),
                method="lm",
            )
            lambas = res.x

            directional_distribution[ipoint, ifreq, :] = mem2_directional_distribution(
                lambas, direction_increment, twiddle_factors
            )

    return directional_distribution


# Numba Implementation
# =============================================================================


# spatial iteration
# ---------------------


# To note; enabling caching seems to not play nice with paralel
@njit(parallel=_PARALLEL, cache=(not _PARALLEL))
def mem2_newton(
    directions_radians: numpy.ndarray,
    a1: numpy.ndarray,
    b1: numpy.ndarray,
    a2: numpy.ndarray,
    b2: numpy.ndarray,
    progress_bar: ProgressBar = None,
    config: NumbaDict = None,
    approximate: bool = False,
) -> numpy.ndarray:
    """
    Return the directional distribution that maximizes Shannon [ - D log(D) ]
    enthrophy constrained by given observed directional moments.

    :param directions_radians: 1d array of wave directions in radians,
    length[number_of_directions]

    :param a1: 1d array of cosine directional moment as function of position and frequency,
        shape = ( number_of_points,number_of_frequencies)

    :param b1: 1d array of sine directional moment as function of position and frequency,
        shape = ( number_of_points,number_of_frequencies)

    :param a2: 1d array of double angle cosine directional moment as function of position and frequency,
        shape = ( number_of_points,number_of_frequencies)

    :param b2: 1d array of double angle sine directional moment as function of position and frequency,
        shape = ( number_of_points,number_of_frequencies)

    :param progress_bar: Progress bar instance if updates are desired.

    :return: array with shape [numbrt_of_points, number_of_frequencies,number_of_direction]
    representing the directional distribution of the waves at each frequency.

    Maximize the enthrophy of the solution with entrophy defined as:

           integrate - D * log(D) over directions

    such that the resulting distribution D reproduces the observed moments.

    """

    number_of_frequencies = a1.shape[-1]
    number_of_points = a1.shape[0]

    directional_distribution = numpy.zeros(
        (number_of_points, number_of_frequencies, len(directions_radians))
    )

    direction_increment_downward_difference = (
        directions_radians - numpy.roll(directions_radians, 1) + numpy.pi
    ) % (2 * numpy.pi) - numpy.pi

    direction_increment_upward_difference = (
        -(directions_radians - numpy.roll(directions_radians, -1) + numpy.pi)
        % (2 * numpy.pi)
        - numpy.pi
    )

    direction_increment = (
        direction_increment_downward_difference + direction_increment_upward_difference
    ) / 2

    # Calculate the needed Fourier transform twiddle factors to calculate moments.
    twiddle_factors = numpy.empty((4, len(directions_radians)))
    twiddle_factors[0, :] = numpy.cos(directions_radians)
    twiddle_factors[1, :] = numpy.sin(directions_radians)
    twiddle_factors[2, :] = numpy.cos(2 * directions_radians)
    twiddle_factors[3, :] = numpy.sin(2 * directions_radians)

    guess = initial_value(a1, b1, a2, b2)
    for ipoint in prange(0, number_of_points):
        if progress_bar is not None:
            progress_bar.update(1)

        # Note; entries to directional_distribution[ipoint, :, :] is modified in the call below. This avoids creation
        # of memory for the resulting array at the expense of allowing for side-effects.
        _mem2_newton_point(
            directional_distribution[ipoint, :, :],
            a1[ipoint, :],
            b1[ipoint, :],
            a2[ipoint, :],
            b2[ipoint, :],
            guess[ipoint, :, :],
            direction_increment,
            twiddle_factors,
            config,
            approximate,
        )

    return directional_distribution


# frequency iteration
# ----------------------


@njit(cache=True)
def _mem2_newton_point(
    out,
    a1,
    b1,
    a2,
    b2,
    guess,
    direction_increment,
    twiddle_factors,
    config=None,
    approximate=False,
):
    """

    :param out: a (view) of the array that will containt the output
    :param a1: 1d array of cosine directional moment as function of frequency,
    :param b1: 1d array of sine directional moment as function of frequency,
    :param a2: 1d array of double angle cosine directional moment as function of frequency,
    :param b2: 1d array of double angle sine directional moment as function of frequency,
    :param guess: initial guess of the lagrange multipliers
    :param direction_increment: directional stepsize used in the integration, nd-array
    :param twiddle_factors: [sin theta, cost theta, sin 2*theta, cos 2*theta] as a 4 by ndir array
    :param config: numerical settings, see description at NUMERICS at top of file.
    :param approximate: whether or not to use the approximate relations.
    :return: None - we use side-effects to pass the results back to the caller (modifying out)
    """
    number_of_frequencies = a1.shape[0]
    for ifreq in range(0, number_of_frequencies):
        #
        moments = numpy.array([a1[ifreq], b1[ifreq], a2[ifreq], b2[ifreq]])
        out[ifreq, :] = mem2_newton_solver(
            moments,
            guess[ifreq, :],
            direction_increment,
            twiddle_factors,
            config,
            approximate,
        )


# mem2 numerical solver
# ----------------------


@njit(cache=True)
def mem2_newton_solver(
    moments: numpy.ndarray,
    guess: numpy.ndarray,
    direction_increment: numpy.ndarray,
    twiddle_factors: numpy.ndarray,
    config=None,
    approximate=False,
) -> numpy.ndarray:
    """
    Newton iteration to find the solution to the non-linear system of constraint equations defining the lagrange
    multipliers in the MEM2 method. Because the Lagrange multipliers enter the equations as exponents the system can
    be unstable to solve numerically.

    :param moments: the normalized directional moments [a1,b1,a2,b2]
    :param guess: first guess for the lagrange multipliers (ndarray, length 4)
    :param direction_increment: directional stepsize used in the integration, nd-array
    :param twiddle_factors: [sin theta, cost theta, sin 2*theta, cos 2*theta] as a 4 by ndir array
    :param config: numerical settings, see description at NUMERICS at top of file.
    :param approximate: whether or not to use the approximate relations.
    :return:
    """
    if config is None:
        max_iter = 100
        rcond = 1e-6
        atol = 0.01
        max_line_search_depth = 8
        use_mem_when_failing_to_converge = True

    else:
        max_iter = config["max_iter"]
        rcond = config["rcond"]
        atol = config["atol"]
        max_line_search_depth = config["max_line_search_depth"]
        use_mem_when_failing_to_converge = (
            config["use_mem_when_failing_to_converge"] > 0.0
        )

    directional_distribution = numpy.empty(len(direction_increment))
    if numpy.any(numpy.isnan(guess)):
        directional_distribution[:] = 0
        return directional_distribution

    if approximate:
        directional_distribution[:] = mem2_directional_distribution(
            guess, direction_increment, twiddle_factors
        )
        return directional_distribution

    current_iterate = guess
    current_func = moment_constraints(
        current_iterate,
        twiddle_factors,
        moments,
        direction_increment,
    )

    jacobian = numpy.empty((4, 4))

    convergence = False
    for iter in range(0, max_iter):

        # Stopping criterium
        magnitude_cur_func_eval = norm(current_func)
        if magnitude_cur_func_eval < atol:
            convergence = True
            break

        #
        # Compute jacobian, and find newton iterate innovation as we solve for:
        #
        #       jacobian @ delta = - current_iterate_func_eval
        #
        # with:
        #
        #       delta = next_lagrange_multiplier_iterate-cur_lagrange_multiplier_iterate

        jacobian = mem2_jacobian(
            current_iterate, twiddle_factors, direction_increment, jacobian
        )
        try:
            update_iterate = solve_cholesky(jacobian, -current_func)
        except Exception:
            update_iterate = numpy.linalg.lstsq(jacobian, -current_func, rcond=rcond)[0]

        magnitude_current_iterate = norm(current_iterate)
        magnitude_update = norm(update_iterate)

        # Do a line search for the optimum decrease. This is intended to stabilize the algorithm
        # as the equations are ill-posed.
        line_search_factor = 1
        for ii in range(max_line_search_depth):
            next_iterate = current_iterate + line_search_factor * update_iterate
            next_func = moment_constraints(
                next_iterate, twiddle_factors, moments, direction_increment
            )

            if norm(next_func) < magnitude_cur_func_eval:
                # If we are decreasing- continue
                current_func = next_func
                current_iterate = next_iterate
                break
            else:
                # The update may be too big as we are not decreasing the cost function magnitude. We will decrease the
                # step size we take - but keep the direction of the step the same.
                inverse_relative_update = magnitude_current_iterate / magnitude_update
                line_search_factor = min(
                    inverse_relative_update, line_search_factor / 2
                )
        else:
            # The linesearch failed. We could not find a factor that ensures the next function estimate is closer
            # to 0.
            convergence = False
            break
    else:
        # We failed to converge after the maximum number of iterations.
        convergence = False

    if not convergence:
        if use_mem_when_failing_to_converge:
            directions = numpy.arctan2(twiddle_factors[1, :], twiddle_factors[0, :])
            directional_distribution[:] = numba_mem(
                directions, moments[0], moments[1], moments[2], moments[3]
            )
        else:
            raise ValueError("we did not converge")

    directional_distribution[:] = mem2_directional_distribution(
        current_iterate, direction_increment, twiddle_factors
    )

    return directional_distribution


# mem2 functions
# ----------------------


@njit(cache=True, fastmath=_FASTMATH)
def moment_constraints(lambdas, twiddle_factors, moments, direction_increment):
    """
    Construct the nonlinear equations we need to solve for lambda. The constrainst are the difference between the
    desired moments a1,b1,a2,b2 and the moment calculated from the current distribution guess and for a perfect fit
    should be 0.

    To note: we differ from Kim et al here who formulate the constraints using unnormalized equations. Here we opt to
    use the normalized version as that allows us to cast the error / or mismatch directly in terms of an error in the
    moments.

    :param lambdas: the lagrange multipliers
    :param twiddle_factors: [sin theta, cost theta, sin 2*theta, cos 2*theta] as a 4 by ndir array
    :param moments: [a1,b1,a2,b2]
    :param direction_increment: directional stepsize used in the integration, nd-array
    :return: array (length=4) with the difference between desired moments and those calculated from the current
        approximate distribution
    """

    # Get the current estimate of the directional distribution
    dist = mem2_directional_distribution(lambdas, direction_increment, twiddle_factors)
    out = numpy.zeros(4)
    for mm in range(0, 4):
        # note - the part after the "-" is just a discrete approximation of the Fourier sine/cosine amplitude (moment)
        out[mm] = moments[mm] - numpy.sum(
            (twiddle_factors[mm, :]) * dist * direction_increment
        )

    return out


@njit(cache=True, fastmath=_FASTMATH)
def mem2_jacobian(lagrange_multiplier, twiddle_factors, direction_increment, jacobian):
    """
    Calculate the jacobian of the constraint equations. The resulting jacobian is a square and positive definite matrix

    :param lambdas: the lagrange multipliers
    :param twiddle_factors: [sin theta, cost theta, sin 2*theta, cos 2*theta] as a 4 by ndir array
    :param direction_increment: directional stepsize used in the integration, nd-array

    :return: a 4 by 4 matrix that is the Jacobian of the constraint equations.
    """
    inner_product = numpy.zeros(twiddle_factors.shape[1])
    for jj in range(0, 4):
        inner_product = inner_product + lagrange_multiplier[jj] * twiddle_factors[jj, :]

    # We subtract the minimum to ensure that the values in the exponent do not become too large. This amounts to
    # multiplyig with a constant - which is fine since we normalize anyway. Effectively- this avoids overflow errors
    # (or infinities) - at the expense of underflowing (which is less of an issue).
    #
    inner_product = inner_product - numpy.min(inner_product)

    normalization = 1 / numpy.sum(numpy.exp(-inner_product) * direction_increment)
    shape = numpy.exp(-inner_product)

    normalization_derivative = numpy.zeros(4)
    for mm in range(0, 4):
        normalization_derivative[mm] = normalization * numpy.sum(
            twiddle_factors[mm, :] * numpy.exp(-inner_product) * direction_increment
        )

    # To note- we have to multiply seperately to avoid potential underflow/overflow errors.
    normalization_derivative = normalization_derivative * normalization

    shape_derivative = numpy.zeros((4, twiddle_factors.shape[1]))
    for mm in range(0, 4):
        shape_derivative[mm, :] = -twiddle_factors[mm, :] * shape

    for mm in range(0, 4):
        # we make use of symmetry and only explicitly calculate up to the diagonal
        for nn in range(0, mm + 1):
            jacobian[mm, nn] = -numpy.sum(
                twiddle_factors[mm, :]
                * direction_increment
                * (
                    normalization * shape_derivative[nn, :]
                    + shape * normalization_derivative[nn]
                ),
                -1,
            )
            if nn != mm:
                jacobian[nn, mm] = jacobian[mm, nn]
    return jacobian


@njit(cache=True, fastmath=_FASTMATH)
def mem2_directional_distribution(
    lagrange_multiplier, direction_increment, twiddle_factors
) -> numpy.ndarray:
    """
    Given the solution for the Lagrange multipliers- reconstruct the directional
    distribution.
    :param lagrange_multiplier: the lagrange multipliers
    :param twiddle_factors: [sin theta, cost theta, sin 2*theta, cos 2*theta] as a 4 by ndir array
    :param direction_increment: directional stepsize used in the integration, nd-array
    :return: Directional distribution arrasy as a function of directions
    """
    inner_product = numpy.zeros(twiddle_factors.shape[1])
    for jj in range(0, 4):
        inner_product = inner_product + lagrange_multiplier[jj] * twiddle_factors[jj, :]

    inner_product = inner_product - numpy.min(inner_product)

    normalization = 1 / numpy.sum(numpy.exp(-inner_product) * direction_increment)
    return numpy.exp(-inner_product) * normalization


@njit(cache=True, fastmath=_FASTMATH)
def initial_value(
    a1: numpy.ndarray, b1: numpy.ndarray, a2: numpy.ndarray, b2: numpy.ndarray
):
    """
    Initial guess of the Lagrange Multipliers according to the "MEM AP2" approximation
    found im Kim1995

    :param a1: moment a1
    :param b1: moment b1
    :param a2: moment a2
    :param b2: moment b2
    :return: initial guess of the lagrange multipliers, with the same leading dimensions as input.
    """
    guess = numpy.empty((*a1.shape, 4))
    fac = 1 + a1**2 + b1**2 + a2**2 + b2**2
    guess[..., 0] = 2 * a1 * a2 + 2 * b1 * b2 - 2 * a1 * fac
    guess[..., 1] = 2 * a1 * b2 - 2 * b1 * a2 - 2 * b1 * fac
    guess[..., 2] = a1**2 - b1**2 - 2 * a2 * fac
    guess[..., 3] = 2 * a1 * b1 - 2 * b2 * fac
    return guess


@njit(cache=True, fastmath=True)
def solve_cholesky(matrix, rhs):
    """
    Solve using cholesky decomposition according to the Cholesky–Banachiewicz algorithm.
    See: https://en.wikipedia.org/wiki/Cholesky_decomposition#The_Cholesky_algorithm
    """
    M, N = matrix.shape
    x = numpy.zeros(M)
    cholesky_decomposition = numpy.zeros((M, M))
    inv = numpy.zeros(M)

    for mm in range(0, M):
        forward_sub_sum = rhs[mm]
        for nn in range(0, mm):
            sum = matrix[mm, nn]
            for kk in range(0, nn):
                sum -= cholesky_decomposition[mm, kk] * cholesky_decomposition[nn, kk]

            cholesky_decomposition[mm, nn] = inv[nn] * sum
            forward_sub_sum += -cholesky_decomposition[mm, nn] * x[nn]

        sum = matrix[mm, mm]
        for kk in range(0, mm):
            sum -= cholesky_decomposition[mm, kk] ** 2

        if sum <= 0.0:
            raise ValueError(
                "Matrix not positive definite, likely due to finite precision errors."
            )

        cholesky_decomposition[mm, mm] = numpy.sqrt(sum)
        inv[mm] = 1 / cholesky_decomposition[mm, mm]
        x[mm] = forward_sub_sum * inv[mm]

    # Backward Substitution (in place)
    for mm in range(0, M):
        kk = M - mm - 1
        sum = x[kk]
        for nn in range(kk + 1, N):
            sum += -cholesky_decomposition[nn, kk] * x[nn]
        x[kk] = sum * inv[kk]
    return x



import numpy
from pandas import DataFrame
from roguewave.interpolate.dataarray import interpolate_track_data_arrray
from roguewave.interpolate.geometry import Geometry, convert_to_track_set
from roguewave.tools.time import to_datetime64
from roguewave.interpolate.nd_interp import NdInterpolator
from typing import Dict, Tuple, Mapping
from xarray import Dataset, DataArray


def interpolate_dataset_grid(
    coordinates,
    data_set: Dataset,
    periodic_data: Mapping[str, Tuple[int, int]] = None,
    longitude_variable_in_dataset: str = "longitude",
) -> Dataset:
    if periodic_data is None:
        periodic_data = {}
        if longitude_variable_in_dataset in data_set:
            periodic_data[longitude_variable_in_dataset] = (360, 180)

        for variable in data_set:
            if "direction" in str(variable).lower():
                periodic_data[variable] = (360, 360)

    first = True
    return_data_set = None
    for coordinate_name, coordinate_value in coordinates.items():
        if first:
            _data_set = data_set
            first = False
        else:
            _data_set = return_data_set
        return_data_set = interpolate_dataset_along_axis(
            coordinate_value, _data_set, coordinate_name
        )
    return return_data_set


def interpolate_dataset_along_axis(
    coordinate_value,
    data_set: Dataset,
    coordinate_name: str = "time",
    periodic_data: Mapping[str, Tuple[int, int]] = None,
    periodic_coordinates: Dict = None,
) -> Dataset:
    if periodic_data is None:
        periodic_data = {"longitude": (360, 180)}
        for variable in data_set:
            if "direction" in str(variable).lower():
                periodic_data[variable] = (360, 360)

    if periodic_coordinates is None:
        periodic_coordinates = {"longitude": 360, "direction": 360}

    if coordinate_name == "time":
        # Ensure that we work with datetime64 for comparisons.
        coordinate_value = to_datetime64(coordinate_value)

    points = {coordinate_name: numpy.atleast_1d(coordinate_value)}
    return_data_set = Dataset()

    for variable in data_set:

        if coordinate_name not in list(data_set[variable].coords):
            return_data_set[variable] = data_set[variable]
            continue

        coordinate_names = [str(x) for x in data_set[variable].dims]
        data_coordinates = [
            (name, data_set[variable].coords[name].values) for name in coordinate_names
        ]

        period_data, discont_data = None, None
        if variable in periodic_data:
            period_data, discont_data = periodic_data[variable]

        dimensions = list(data_set[variable].dims)

        def get_data(indices, idims):
            index = [slice(None)] * len(dimensions)
            for interp_index, idim in zip(indices, idims):
                index[idim] = interp_index
            return data_set[variable].values[tuple(index)]

        data_accessing = NdInterpolator(
            get_data,
            data_coordinates,
            data_set[variable].shape,
            list(points.keys()),
            coordinate_name,
            periodic_coordinates,
            period_data,
            discont_data,
        )

        coor = {value[0]: value[1] for value in data_coordinates}
        for coor_name, coor_value in points.items():
            coor[coor_name] = coor_value

        return_data_set[variable] = DataArray(
            data=data_accessing.interpolate(points),
            coords=coor,
            dims=data_set[variable].dims,
        )

    return return_data_set


def interpolate_dataset(
    data_set: Dataset,
    geometry: Geometry,
    periodic_coordinates: Dict[str, float] = None,
    periodic_data: Dict[str, Tuple[float, float]] = None,
    time_variable_in_dataset: str = "time",
    longitude_variable_in_dataset: str = "longitude",
    latitude_variable_in_dataset: str = "latitude",
):
    geometry = convert_to_track_set(geometry, data_set[time_variable_in_dataset].values)

    if periodic_coordinates is None:
        periodic_coordinates = {longitude_variable_in_dataset: 360}

    for variable in data_set:
        if "direction" in str(variable).lower():
            periodic_data = {variable: (360, 360)}

    out = {}
    for name, track in geometry.tracks.items():
        points = {
            time_variable_in_dataset: track.time,
            longitude_variable_in_dataset: track.longitude,
            latitude_variable_in_dataset: track.latitude,
        }
        out[name] = interpolate_at_points(
            data_set=data_set,
            points=points,
            independent_variable=time_variable_in_dataset,
            periodic_coordinates=periodic_coordinates,
            periodic_data=periodic_data,
        ).to_dataframe()
        if out[name].index.tz is None:
            out[name].index = out[name].index.tz_localize("utc")
        out[name].reset_index(inplace=True)
    return out


def interpolate_at_points(
    data_set: Dataset,
    points: Dict[str, numpy.ndarray],
    independent_variable=None,
    periodic_coordinates: Dict[str, float] = None,
    periodic_data: Dict[str, Tuple[float, float]] = None,
) -> Dataset:
    if periodic_data is None:
        periodic_data = {}

    dimensions = data_set.dims
    if independent_variable is None:
        if "time" in dimensions:
            independent_variable = "time"
        else:
            independent_variable = dimensions[0]

    return_data_set = Dataset(
        coords={independent_variable: points[independent_variable]}
    )
    for variable in data_set:
        if variable in periodic_data:
            period_data = periodic_data[variable][0]
            discont = periodic_data[variable][1]
        else:
            period_data = None
            discont = None

        return_data_set[variable] = interpolate_track_data_arrray(
            data_set[variable],
            points,
            independent_variable,
            periodic_coordinates=periodic_coordinates,
            period_data=period_data,
            discont=discont,
        )
    return return_data_set


def tracks_as_dataset(time, drifter_tracks: Mapping[str, DataFrame]) -> DataArray:
    tracks = {}
    for track_id, track in drifter_tracks.items():

        track.set_index("time", inplace=True)
        variables = list(track.columns)
        tracks[track_id] = DataArray(
            track,
            coords={"time": time, "variables": variables},
            dims=["time", "variables"],
            name=track_id,
        )

    dataset = Dataset(tracks).to_array(dim="spotter_id")
    return dataset



"""
Contents: Routines to get data from the spotter api

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines to get data from the spotter api

Functions:

- `get_spectrum`, function to download spectral data.
- `get_bulk_wave_data`, function to download bulk wave data.
- `get_data`, general function to download different data types (spectral,
    bulk wave, wind, SST, barometer).
- `search_circle`, get all available data within a given circle.
- `search_rectangle`, get all available data within a given rectangle.

"""

# 1) Imports
# =============================================================================
from datetime import datetime, timedelta, timezone
from .exceptions import ExceptionNoDataForVariable
from pysofar.spotter import Spotter, SofarApi
from pysofar.wavefleet_exceptions import QueryError
from roguewave import logger
from roguewave.tools.time import (
    datetime_to_iso_time_string,
    to_datetime_utc,
)
from roguewave.wavespectra import (
    FrequencySpectrum,
    concatenate_spectra,
)

from typing import Dict, List, Union, Sequence, Literal, Any
from pandas import DataFrame, concat
from .helper_functions import (
    _get_sofar_api,
    get_spotter_ids,
    _unique_filter,
    as_dataframe,
    _none_filter,
    _get_class,
)
import roguewave.spotterapi.spotter_cache as spotter_cache

# 2) Constants & Private Variables
# =============================================================================
# Maximum number of spectra to retrieve from the Spotter API per API call. Note
# that 2- os a hard limit of the API. If set higher than 100 it will just
# return 100 (and the implementation will fail)
MAX_LOCAL_LIMIT = 100
MAX_LOCAL_LIMIT_BULK = 500

# Number of retry attemps if a download fails.
NUMBER_OF_RETRIES = 2

MAX_DAYS_SMARTMOORING = 10

DATA_TYPES = Literal[
    "waves",
    "wind",
    "surfaceTemp",
    "barometerData",
    "frequencyData",
    "microphoneData",
    "smartMooringData",
]


# 4) Main Functions
# =============================================================================


def get_spectrum(
    spotter_ids: Union[str, Sequence[str]],
    start_date: Union[datetime, int, float, str] = None,
    end_date: Union[datetime, int, float, str] = None,
    **kwargs,
) -> Dict[str, FrequencySpectrum]:
    """
    Gets the requested frequency wave data for the spotter(s) in the given
    interval.

    :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
    Spotter_id.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to end of spotter history

    :return: Data as a dictornary with spotter_id's as keys, and for each
    corresponding value a List that for each returned time contains a
    WaveSpectrum1D object.

    """
    return get_spotter_data(
        spotter_ids, "frequencyData", start_date, end_date, **kwargs
    )


def get_bulk_wave_data(
    spotter_ids: Union[str, Sequence[str]],
    start_date: Union[datetime, int, float, str] = None,
    end_date: Union[datetime, int, float, str] = None,
    **kwargs,
) -> Dict[str, DataFrame]:
    """
    Gets the requested bulk wave data for the spotter(s) in the given interval

    :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
    Spotter_id.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to end of spotter history

    :return: Data as a dictornary with spotter_id's as keys, and for each
    corresponding value a dataframe containing the output.
    """
    df = get_spotter_data(spotter_ids, "waves", start_date, end_date, **kwargs)
    return {key: value for (key, value) in df.groupby(["spotter_id"])}


# -----------------------------------------------------------------------------
def get_data(
    spotter_ids: Union[str, List[str]],
    start_date: Union[datetime, int, float, str] = None,
    end_date: Union[datetime, int, float, str] = None,
    include_frequency_data=False,
    include_waves=True,
    include_wind=False,
    include_barometer_data=False,
    include_surface_temp_data=False,
    **kwargs,
) -> Dict[str, Dict[str, Union[FrequencySpectrum, DataFrame]]]:
    """
    DEPRICATED USE get_spotter_data instead
    """
    from warnings import warn

    warn(
        "The get_data function is depricated, please use get_spotter_data instead",
        DeprecationWarning,
        stacklevel=2,
    )

    data_to_get = []

    if include_frequency_data:
        data_to_get.append("frequencyData")
    if include_barometer_data:
        data_to_get.append("barometerData")
    if include_surface_temp_data:
        data_to_get.append("surfaceTemp")
    if include_wind:
        data_to_get.append("wind")
    if include_waves:
        data_to_get.append("waves")
    data_to_get: List[DATA_TYPES]

    data = {}
    for data_type in data_to_get:
        df = get_spotter_data(spotter_ids, data_type, start_date, end_date, **kwargs)
        if not data_type == "frequencyData":
            df = {key: value for (key, value) in df.groupby(["spotter_id"])}

        for spotter_id in df:
            if spotter_id not in data:
                data[spotter_id] = {}
            data[spotter_id][data_type] = df[spotter_id]

    return data


# Return types for get spotter data. Options that control the return type are:
# - data-to-get-entries; if frequencyData is requested we return a FrequencySpectrum object,
#                        otherwise we return a dataframe.
# - data-to_get type; if a list we return each requested variable as a key value mapping
# - flatten: if false data for each spotter is returned in a key value mapping with
#            spotter_id as key.


def get_spotter_data(
    spotter_ids: Union[str, Sequence[str]],
    data_type: DATA_TYPES,
    start_date: Union[datetime, int, float, str] = None,
    end_date: Union[datetime, int, float, str] = None,
    session: SofarApi = None,
    parallel_download=True,
    cache=True,
) -> Union[DataFrame, Dict[str, FrequencySpectrum]]:
    """
    Gets the requested data for the spotter(s) in the given interval as either a dataframe containing
    all the data for the combined spotters in a single table (all datatypes except frequencyData) or
    a dictionary object that has the spotter_id as key and contains a frequency spectrum object as
    values.

    :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
    Spotter_id.

    :param data_type: Literal string denoting the desired data type, options are
            data_type="waves", bulk wave data
            data_type="wind", wind estimates
            data_type="surfaceTemp", surface temperature (if available)
            data_type="barometerData", barometer data (if available)
            data_type="frequencyData", frequency data (if available) NOTE: does not return a datafrae
            data_type="microphoneData", microphone data if available
            data_type="smartMooringData", smartmooring data if available.

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to end of spotter history

    :param session:    Active SofarApi session. If none is provided one will be
                       created automatically. This requires that an API key is
                       set in the environment.

    :param parallel_download: Use multiple requests to the Api to speed up
                       retrieving data. Only useful for large requests.

    :param cache: Cache requests. If True, returned data will be stored in
                        a file Cache on disk, and repeated calls with the
                        same arguments will use locally cached data. The cache
                        is a FileCache with a maximum of 2GB by default.

    :return:
        data_type="frequencyData": a dictionary with spoter ids as keys and FrequencySpectra as values
        data_type= ...  : a Pandas Dataframe with a spotter_id column that indicates to which spotter entries
            belong.
    """

    if spotter_ids is None:
        spotter_ids = get_spotter_ids()

    if session is None:
        session = _get_sofar_api()

    # Make sure we have a list object
    if not isinstance(spotter_ids, (list, tuple)):
        spotter_ids = [spotter_ids]

    # Ensure we have datetime aware objects
    start_date = to_datetime_utc(start_date)
    end_date = to_datetime_utc(end_date)

    description = f"Downloading from api - {data_type}"
    if not cache or end_date > datetime.now(tz=timezone.utc):
        # Use cached data  _only_ for requests that concern the past. In real-time
        # things may change. In the later case (or when we do not want to use cached values) we flush the cache entries.
        spotter_cache.flush(
            spotter_ids,
            session,
            _download_data,
            var_name=data_type,
            start_date=start_date,
            end_date=end_date,
        )

    data_for_variable = spotter_cache.get_data(
        spotter_ids,
        session,
        _download_data,
        parallel=parallel_download,
        description=description,
        var_name=data_type,
        start_date=start_date,
        end_date=end_date,
    )

    if data_type == "frequencyData":
        data = {}
        for spotter_id, spotter_data in zip(spotter_ids, data_for_variable):
            #
            # Did we get any data for this spotter
            if spotter_data is not None:
                data[spotter_id] = spotter_data
    else:
        values, _id = [], []
        for spotter_id, spotter_data in zip(spotter_ids, data_for_variable):
            #
            # Did we get any data for this spotter
            if spotter_data is not None:
                values += [spotter_data]
                _id += [spotter_id]

        data = concat(values, keys=_id, names=["spotter_id", "time index"])

        data.reset_index(inplace=True)
        data.drop(columns="time index", inplace=True)
    return data


def _download_data(**kwargs) -> Union[FrequencySpectrum, DataFrame]:
    """ """
    data = list(_unpaginate(**kwargs))

    # Postprocessing
    if len(data) < 1:
        data = None
    else:
        if kwargs["var_name"] == "frequencyData":
            data = _unique_filter(data)
            data = concatenate_spectra(data, dim="time")
        else:
            data = as_dataframe(data)
            data.drop_duplicates(inplace=True)
            data.sort_values("time", inplace=True)
    return data


def get_smart_mooring_data(
    start_date: str, end_date: str, spotter: Spotter, max_days=None
):

    # We have a self imposed limit that we return to avoid overloading wavefleet.
    if max_days is not None:
        start_date_dt = to_datetime_utc(start_date)
        end_date_dt = to_datetime_utc(end_date)

        # If the enddate exceeds the maximum number of days to download we reduce the enddate
        if end_date_dt - start_date_dt > timedelta(days=max_days):
            end_date_dt = start_date_dt + timedelta(days=max_days)

            # update the enddate
            end_date = datetime_to_iso_time_string(end_date_dt)

    params = {"spotterId": spotter.id, "startDate": start_date, "endDate": end_date}
    scode, data = spotter._session._get("sensor-data", params=params)

    if scode != 200:
        raise QueryError(data["message"])

    return data["data"]


def _unpaginate(
    var_name: str,
    start_date: datetime,
    end_date: datetime,
    spotter_id: str,
    session: SofarApi,
) -> Union[FrequencySpectrum, Dict[str, Any]]:
    """
    Generator function to unpaginate data from the api.

    idiosyncrasies to handle:
    - Wavefleet sometimes returns multiple instances of the same record (same
      time). These are filtered through a all to _unique. This should
      be fixed in the future.
    - Some entries are broken (None for entries).
    - Not all Spotters will have (all) data for the given timerange.
    - Wavefleet sometimes times out on requests.

    :param spotter_id: Spotter id

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters
                       history

    :param end_date: ISO 8601 formatted date string, epoch or datetime.
                     If not included defaults to end of spotter history

    :return: Data
    """
    spotter = Spotter(spotter_id, spotter_id, session=session)
    page = 0
    while True:
        json_data = None
        page += 1
        for retry in range(0, NUMBER_OF_RETRIES + 1):
            try:
                if var_name == "smartMooringData":
                    json_data = get_smart_mooring_data(
                        start_date=datetime_to_iso_time_string(start_date),
                        end_date=datetime_to_iso_time_string(end_date),
                        spotter=spotter,
                        max_days=MAX_DAYS_SMARTMOORING,
                    )
                    json_data = {"smartMooringData": json_data}
                else:
                    json_data = spotter.grab_data(
                        limit=MAX_LOCAL_LIMIT,
                        start_date=datetime_to_iso_time_string(start_date),
                        end_date=datetime_to_iso_time_string(end_date),
                        include_frequency_data=var_name == "frequencyData",
                        include_directional_moments=var_name == "frequencyData",
                        include_barometer_data=var_name == "barometerData",
                        include_waves=var_name == "waves",
                        include_wind=var_name == "wind",
                        include_surface_temp_data=var_name == "surfaceTemp",
                        include_microphone_data=var_name == "microphoneData",
                    )
                break
            except Exception as e:
                if retry < NUMBER_OF_RETRIES:
                    warning = (
                        f"Error downloading data for {spotter.id}, "
                        f"attempting retry {retry + 1}"
                    )
                    logger.warning(warning)
                else:
                    raise e

        #
        # If so- was it returned? If not- raise error
        if var_name not in json_data:
            raise ExceptionNoDataForVariable(var_name)

        # If no data - return
        if not json_data[var_name]:
            break

        # Filter for Nones.
        json_data[var_name] = _none_filter(json_data[var_name])
        if not json_data[var_name]:
            break

        objects = []
        for _object in [_get_class(var_name, data) for data in json_data[var_name]]:
            if isinstance(_object, FrequencySpectrum):
                date = to_datetime_utc(_object.time.values)
            else:
                date = _object["time"]

            if date < start_date:
                continue
            else:
                objects.append(_object)

        if len(objects) == 0:
            break

        last_object = None
        for _object in objects:
            last_object = _object
            yield _object

        if isinstance(last_object, FrequencySpectrum):
            start_date = to_datetime_utc(last_object.time.values) + timedelta(seconds=1)
        else:
            start_date = last_object["time"] + timedelta(seconds=1)



# Import
# =============================================================================
from datetime import datetime, timedelta

from .modelinformation import _get_resource_specification, \
    _RemoteResourceSpecification, _find_between
from .timebase import TimeSlice
from typing import List


# Main Public Functions
# =============================================================================
def generate_uris(
        variable: str,
        time_slice: TimeSlice,
        model_name: str) -> List[str]:
    """
    Get the AWS keys associated with a time slice for a given
    model.

    :param variable: name of the variable of interest
    :param time_slice: how we slice the forecast time
    :param model_name: model name
    :return: List of valid uris.
    """
    resource = _get_resource_specification(model_name)
    time_vector = time_slice.time_base(resource.model_time_configuration)

    aws_keys = []
    for time in time_vector:
        aws_keys.append(
            _generate_aws_key(
                variable=variable,
                init_time=time[0],
                forecast_hour=time[1],
                model=resource)
        )
    return aws_keys


def _generate_aws_key(variable, init_time: datetime, forecast_hour: timedelta,
                      model: _RemoteResourceSpecification) -> str:
    """
    Function that generates a valid AWS key for the given model that represents
    date from the model initialized at the given init_time and with the given
    forecast_hour or lead time.

    :param variable: variable name we want to retrieve (as used in key, if
        applicable, set to None otherwise)
    :param init_time: init time of the model we want to retrieve
    :param forecast_hour: the forecast hour of the model we want to retrieve
    :param model: Model() information on where data is stored.
    :return: A valid aws key for the given model.
    """

    # Helper functions. These are local for now and tied to this function
    # alone. Hence the encapsulation.
    def _replace(_string: str, key: str, value: str):
        """
        Replace a key in string with value
        :param _string: target
        :param key:  string to replace
        :param value:  string to replace with
        :return: string
        """
        return _string.replace("{" + key + "}", value)

    #

    def _time(_string, key, value: datetime):
        """
        Replace {key:fmt} with value.strftime(fmt)
        :param _string: target
        :param key:  either "init_time" or "valid_time"
        :param value: init_time or valid_time as datetime
        :return: string
        """

        # There may be multiple instances that need to be replaced:
        while key in _string:
            # Get the time format string
            _format = _find_between(_string, '{' + key + ':', '}')
            # get the text we want to replace
            to_replace = key + ':' + _format
            # what we want to replace it with (the formatted time)
            replace_with = value.strftime(_format)
            # replace
            _string = _replace(_string, to_replace, replace_with)
        return _string

    #

    def _ecmwf_special(_string, key, _init_time: datetime,
                       _forecast_hour: timedelta):
        """
        Special parsing only valid for ecmwf formatted output.
        :param _string: string
        :param key:  ecmwf_subcycle_string or ecmwf_subcycle_string
        :param _init_time: init time as datetime
        :param _forecast_hour: forecast hour as timedelta
        :return:
        """

        if key == 'ecmwf_lead_zero_indicator':
            # the suffix of ecmwf forecasts is either `001` for all forecast
            # hours exept the zero hour.
            if _forecast_hour == timedelta(hours=0):
                ecmwf_lead_zero_indicator = '011'
            else:
                ecmwf_lead_zero_indicator = '001'
            return _replace(_string, key, ecmwf_lead_zero_indicator)

        elif key == 'ecmwf_subcycle_string':
            # the prefix of ecmwf forecasts is A2P for 10 day forecasts (00Z
            # and # 12Z cycles), or A2Q for the 4 day forecasts (06Z and 18Z
            # cycles)
            if (_init_time.hour == 0) or (_init_time.hour == 12):
                ecmwf_subcycle_string = 'A2P'
            else:
                ecmwf_subcycle_string = 'A2Q'
            return _replace(_string, key, ecmwf_subcycle_string)
        else:
            raise Exception(f'Unknown key {key}')

    # ---------------------

    #
    # Purpose:
    #
    #   The target template key has the form
    #   "t/{init_time:%Y%m%d}/{init_time:%H}/t/{forecasthour}.{variable}.nc"
    #   here everything between {...} defines a token that needs to be replaced
    #   at executing time by a value. E.g. {init_time:%Y%m%d} denotes we want
    #   to replace this with the init_time formated according to %Y%m%d. This
    #   function does the parsing/replacement.
    #
    # Implementation:
    #
    #  We define a set of properties below, where the key denotes the keyword
    #  we want to replace (init_time, forecast_hour etc.), and the value is
    #  a dictionary with a method denoting the function handling the
    #  replacement and a set of arguments in addition to the string and the key
    #  to be replaced.
    valid_time = init_time + forecast_hour

    # Define keywords
    keywords = {
        "forecasthour": {
            'method': _replace, 'args': (
                "{hours:03d}".format(
                    hours=int(forecast_hour.total_seconds() // 3600)
                ),
            )
        },
        "init_time": {'method': _time, 'args': (init_time,)},
        "valid_time": {'method': _time, 'args': (valid_time,)},
        "variable": {'method': _replace, 'args': (variable,)},
        "ecmwf_subcycle_string": {
            'method': _ecmwf_special, 'args': (init_time, forecast_hour)
        },
        "ecmwf_lead_zero_indicator": {
            'method': _ecmwf_special, 'args': (init_time, forecast_hour)
        }
    }
    # Loop over properties and invoke method to replace keywords.
    uri = model.uri_path_template(variable)
    for keyword, action in keywords.items():
        if keyword in uri:
            uri = action['method'](uri, keyword, *action['args'])

    return uri


"""
This module is primarily intended to read model setup data from a model
definition file generated by WW3 (v6.07). There exists (to my knowledge) no
formal definition of a ww3 model_def file - as the format is primarily intended
to be used by ww3 internally. Here we need the information in the model_def
file so we can read restart files. The current implementation was generated
from reverse engineering from the Fortran source how we should interpret the
binary, unformatted sequential file (Fortran writing type). We do _not_ try
and read all data (some is ignored)- only an effort is made to read what we
need for restart files.

To note:
In case of changes in ww3 that change the format it may be easier to instead
of rewriting an interpreter just providing the needed information directly.
Specifically we need:
- frequencies
- directions
- depth
- relation linear index to ilat, ilon
- relation ilat, ilon to linear index
- number of "sea" or spatial points.

All of these can in principle be inferred form the way the model is set-up. We
do not _need_ to read the model_def file (though it is convinient to do this
automatically).

Pieter Smit, Aug, 2022

"""

import numpy

from roguewave.wavewatch3.grid_tools import Grid
from roguewave.wavewatch3.fortran_types import (
    FortranInt,
    FortranFloat,
    FortranCharacter,
)
from roguewave.wavewatch3.resources import Resource
from io import BytesIO
from typing import Union, Literal


def read_model_definition(
    resource: Resource, byte_order: Literal["<", ">", "="] = "<"
) -> Grid:

    """
    Read a ww3 model definition file, see comments on top of this file for
    further info.

    :param resource: an instance of a resource.
    :param byte_order: which byte order data is stored. Sofar currently uses
        little endian (<).
    :return: the grid, depth and the landmask.
    """

    def jump(stream, number, start):
        if start is None:
            _ = stream.read(number)
        else:
            stream.seek(start + number)

    stream = BytesIO(resource.read())

    fort_char = FortranCharacter(endianness=byte_order)
    fort_int = FortranInt(endianness=byte_order)
    fort_float = FortranFloat(endianness=byte_order)

    data = {}
    start_of_record_absolute = 0

    # The need to read the start of record (_sor) and end of record (_eor) is
    # a quirk of the sequential unformatted IO that ww3 uses to write this
    # file and our use of gfortran to compile. Specifically:
    #
    # https://gcc.gnu.org/
    #    onlinedocs/gfortran/File-format-of-unformatted-sequential-files.html
    #
    # Each record in the file contains a leading
    # start of record (sor) 4 byte marker indicating size N of the record,
    # the data written (N bytes), and a trailing end of record (eor) marker again
    # indicating the size of the data. No information on the type of the data is
    # provided, and this has to be known in advance. E.g. mulitple different
    # information types can (and are) stored as part of the same record.
    _sor = fort_int.unpack(stream, 1)[0]  # Fortran record opening
    data["id"] = fort_char.unpack(stream, 35)  # Line 70
    data["grid_name"] = fort_char.unpack(stream, 10)  # Line 70  # noqa: F841
    data["nlon"] = fort_int.unpack(stream, 1)[0]  # Number of longitude points
    data["nlat"] = fort_int.unpack(stream, 1)[0]  # Number of longitude points
    data["num_sea"] = fort_int.unpack(stream, 1)[0]  # Number of sea points
    data["num_dir"] = fort_int.unpack(stream, 1)[0]  # Number of directional points
    data["num_freq"] = fort_int.unpack(stream, 1)[0]  # Number of wavenumber points

    # Number of input bound points (see w3odatd.ftn, line 219) and
    # Number of files for output bound data (see w3odatd.ftn, line 220)
    # (3 int total, not used here, skipping, jumping to eor
    _ = jump(stream, _sor + 4, 0)
    _eor = fort_int.unpack(stream, 1)[0]
    assert _sor == _eor

    _sor = fort_int.unpack(stream, 1)[0]
    _ = jump(stream, _sor, None)
    _eor = fort_int.unpack(stream, 1)[0]
    assert _sor == _eor

    # assert fort_int.unpack(stream,1)[0] == _sor

    # Enter the section for W3GDAT (line 582)
    data["grid_type"], data["flagll"], data["iclose"] = fort_int.unpack(
        stream, None, unformatted_sequential=True
    )
    (
        data["longitude_stepsize"],
        data["latitude_stepsize"],
        data["longitude_start"],
        data["latitude_start"],
    ) = fort_float.unpack(stream, None, unformatted_sequential=True)

    _sor = fort_int.unpack(stream, 1)[0]
    start_of_record_absolute = stream.tell()
    dtype_float = numpy.dtype("float32").newbyteorder(byte_order)
    dtype_int = numpy.dtype("int32").newbyteorder(byte_order)

    # First the bottom grid. The bottom grid is only stored on the
    # computational grid
    data["bottom_datum"] = numpy.frombuffer(
        stream.read(data["num_sea"] * 4), count=data["num_sea"], dtype=dtype_float
    )

    # Next the mask layer. The masked layer is stored everywhere. Note that it
    # is stored with latitude as the fast axis- here we retrieve as longitude
    # as the fast axis (transpose)
    num_points = data["nlon"] * data["nlat"]
    data["mask"] = numpy.reshape(
        numpy.frombuffer(stream.read(num_points * 4), dtype=dtype_int),
        (data["nlon"], data["nlat"]),
    ).transpose()

    # Next the mapping that relates ix,iy -> ns. Subtract 1 to account for 0
    # based indexing (vs 1 in Fortran). Note that it is stored with latitude as
    # the fast axis- here we retrieve as longitude as the fast axis (transpose)
    data["to_linear_index"] = (
        numpy.reshape(
            numpy.frombuffer(stream.read(num_points * 4), dtype=dtype_int),
            (data["nlon"], data["nlat"]),
        ).transpose()
        - 1
    )

    num_points = data["num_sea"]
    data["to_point_index"] = (
        numpy.reshape(
            numpy.frombuffer(stream.read(num_points * 8), dtype=dtype_int),
            (2, num_points),
        )
        - 1
    )

    # FROM: w3gdatmd.ftn line 160
    #      TRFLAG    Int.  Public   Flag for use of transparencies
    #                                0: No sub-grid obstacles.
    #                                1: Obstructions at cell boundaries.
    #                                2: Obstructions at cell centers.
    #                                3: Like 1 with continuous ice.
    #                                4: Like 2 with continuous ice.
    data["tr_flag"] = fort_int.unpack(stream, 1)[0]
    _ = jump(stream, _sor, start_of_record_absolute)
    _eor = fort_int.unpack(stream, 1)[0]
    assert _sor == _eor

    # we jump three values associated with TRFLAG
    for ii in range(0, 3):
        _ = fort_int.unpack(stream, None, True)

    # Spectral parameters
    # For descriptions; see w3gdatmd.ftn
    num_spec_points = data["num_freq"] * data["num_dir"]

    # I do not fully understand why arrays we ignore are larger.. but we
    # ignore them
    byte_size_full = (num_spec_points + data["num_dir"]) * 4

    # ignore: "mapwn, mapth"
    _sor = fort_int.unpack(stream, 1)[0]
    _ = stream.read(byte_size_full * 2)
    data["direction_step_size"] = fort_float.unpack(stream, 1)[0]
    data["direction_degree"] = (
        numpy.frombuffer(stream.read(data["num_dir"] * 4), dtype=dtype_float)
        * 180
        / numpy.pi
    )

    # ignore: "stuff"
    _ = stream.read(byte_size_full * 5)

    #
    data["freq_mult_fac"] = fort_float.unpack(stream, 1)[0]
    data["start_frequency"] = fort_float.unpack(stream, 1)[0]

    # Note they store an extra frequency before start, and at the end, I
    # assume this is for convenience in calculating delta's- either way,
    # ignored here.
    data["frequency_hertz"] = (
        numpy.frombuffer(stream.read(data["num_freq"] * 4 + 8), dtype=dtype_float)
        / numpy.pi
        / 2
    )[1:-1]

    latitude = numpy.linspace(
        data["latitude_start"],
        data["latitude_start"] + data["latitude_stepsize"] * (data["nlat"] - 1),
        data["nlat"],
        endpoint=True,
    )

    longitude = numpy.linspace(
        data["longitude_start"],
        data["longitude_start"] + data["longitude_stepsize"] * data["nlon"],
        data["nlon"],
        endpoint=False,
    )
    depth = -data["bottom_datum"]
    grid = Grid(
        number_of_spatial_points=data["num_sea"],
        frequencies=data["frequency_hertz"],
        directions=data["direction_degree"],
        latitude=latitude,
        longitude=longitude,
        depth=depth,
        mask=data["mask"],
        _to_linear_index=data["to_linear_index"],
        _to_point_index=data["to_point_index"],
    )

    return grid


def _to_slice(val: Union[slice, int]) -> slice:
    if not isinstance(val, (slice, int)):
        raise ValueError("Only slice or int supported as index.")

    if isinstance(val, int):
        return slice(val, val + 1, 1)
    else:
        return val



from typing import Dict
import numpy
from xarray import DataArray

from roguewave.interpolate.nd_interp import NdInterpolator


def interpolate_track_data_arrray(
    data_array: DataArray,
    tracks: Dict[str, numpy.ndarray],
    independent_variable=None,
    periodic_coordinates: Dict[str, float] = None,
    period_data=None,
    discont=None,
) -> DataArray:
    """
    Interpolate a data array from a dataset at given points in N-dimensions.

    The interpolation function takes care of the following edge cases not
    covered by standard xarray interpolation:

    1) Interpolation along cyclic dimensions (e.g. Longitude interpolating
       across prime- or anti-meridian)
    2) Interpolation of cyclic data. E.g. directions.
    3) Interpolation near non-existing values (e.g. NaN values representing
       land).

    When these are covered by standard xarray interpolation we should
    definitely switch.

    :param data_array: xarray data array.

    :param _points: dictionary with data_array coordinate names as keys and
                    a numpy array with N entries. With coordiantes lat/lon:
                    {
                        lat: [ lat0, lat1 ,lat2, .., 10]
                        lon: [ [lon0 ,lon1,lon2, .., lonN]
                        time: [t0,t1,t2, ... tN]
                    }
                    where the points we are interolating are formed by the
                    tuples [  (t[0],lat[0],lon[0]),
                                 ....
                              (t[1],lat[N],lon[N]),
                    ]

    :param independent_variable: coordinate that is used to parametrize the
        points, usually time. This variable is used as the output coordinate
        for the returned data array. If None and 'time' is a valid coordinate
        'time' is used, otherwise it defaults to whatever the first returned
        coordinate is from data_array.dims

    :param periodic_coordinates: Dictonary that contains as keys the coordinate
        names of cyclic coordinates, and as value the cyclic length of the
        coordinate.

    :param period_data: Cyclic length of the data. If data is not cyclic
        this is set to None (default).

    :param fractional_cyclic_range: range of the cyclic data expressed as a
        fraciton of the cyclic lenght. E.g for angles in [-pi,pi] with
        cyclic length 2*pi we would give [-0.5,0.5]. Defaults to [0,1].

    :return: A data array of interpolated values with as coordinates the
        specified independent coordinate (usually time).
    """

    # Get dimensions of the data array
    dimensions = data_array.dims

    if periodic_coordinates is not None:
        for wrapped_coordinate in periodic_coordinates:
            if wrapped_coordinate not in dimensions:
                raise ValueError(
                    f"Cyclic coordinate {wrapped_coordinate} is"
                    f" not a valid coordinate of the dataset."
                )
    else:
        periodic_coordinates = {}

    if independent_variable is None:
        if "time" in dimensions:
            independent_variable = "time"
        else:
            independent_variable = dimensions[0]

    # Ensure that each of the coordinate lists indicated in points is at least
    # 1D
    for coordinate_name in tracks:
        tracks[coordinate_name] = numpy.atleast_1d(tracks[coordinate_name])

    if len(tracks) != len(dimensions):
        dim_str = [str(dim) for dim in dimensions]
        raise ValueError(
            f" Points expected to have {len(dimensions)} "
            f'coordinates corresponding to: {", ".join(dim_str)}'
        )

    for coordinate_name in tracks:
        dim_str = [str(dim) for dim in dimensions]
        if coordinate_name not in dim_str:
            raise ValueError(
                f" Coordinate {coordinate_name} not in data array"
                f'coordinates: {", ".join(dim_str)}'
            )

    coordinates = []
    for index, coordinate_name in enumerate(data_array.dims):
        # Get weights and indices
        coordinates.append((coordinate_name, data_array[coordinate_name].values))

    def _get_data(indices, _dummy):
        return data_array[tuple([DataArray(x) for x in indices])].values

    interpolator = NdInterpolator(
        get_data=_get_data,
        data_coordinates=coordinates,
        data_shape=data_array.shape,
        interp_coord_names=list(tracks.keys()),
        interp_index_coord_name=independent_variable,
        data_periodic_coordinates=periodic_coordinates,
        data_period=period_data,
        data_discont=discont,
    )
    interpolated_values = interpolator.interpolate(tracks)

    return DataArray(
        interpolated_values,
        coords={independent_variable: tracks[independent_variable]},
        dims=independent_variable,
        name=data_array.name,
    )
